/*
 * Jasq platform for JavaScript
 * http://lmadhavan.com/software/jasq/
 *
 * Copyright 2010 L. Madhavan
 * Licensed under the BSD license.
 * See the license.txt file for licensing terms.
 *
 * Last modified on 31 August 2010
 */

/* 'package' and 'import' equivalents for JavaScript */

function package2(name) {
	var a = name.split('.'),
	    base = this;
	name = null;
	for (var i = 0; i < a.length; i++) {
		var p = a[i];
		name = name == null ? a[i] : name + '.' + a[i];
		if (!(base[p] instanceof jasq._Package))
			base[p] = new jasq._Package(name);
		base = base[p];
	}
}

function import2(name) {
	var rn = jasq._resolveName(name),
	    base = rn.base,
	    target = rn.target;
	
	if (target == '*') {
		for (name in base)
			if (base.hasOwnProperty(name)) {
				var clazz = base[name];
				if (!(clazz instanceof jasq._Package))
					this[name] = clazz;
			}
	} else {
		var what = base[target];
		if (typeof what == 'undefined')
			throw new ReferenceError('Invalid import: ' + name);
		else
			this[target] = what;
	}
}

/* Java language features */

jasq = {
	_Class: function(name, clazz, superclazz) {
		this.name = name;
		this.clazz = clazz;
		this.superclazz = superclazz;
	},

	_Package: function(name) {
		this.name = name;
	},

	_resolveName: function(name) {
		var a = name.split('.'),
		    base = (function() { return this; })();

		for (var i = 0; i < a.length - 1; i++) {
			var p = a[i];
			if (!(base[p] instanceof jasq._Package)) {
				a.length = i + 1;
				throw new ReferenceError('Cannot find package: ' + a.join('.'));
			}
			base = base[p];
		}

		return { base: base, target: a[a.length - 1] };
	},

	_inject: function(clazz, properties) {
		var target = clazz.prototype;
		for (p in properties)
			if (properties.hasOwnProperty(p)) {
				var prop = properties[p];
				target[p] = prop;
				if (typeof prop == 'function')
					prop._method = { clazz: clazz.class2, name: p };
			}
		/*@cc_on
		if (properties.hasOwnProperty('toString'))
			target.toString = properties.toString;
		@*/ // Fix for IE
	},

	_injectStatic: function(clazz, properties) {
		var target = clazz;
		for (p in properties)
			if (properties.hasOwnProperty(p)) {
				var prop = properties[p];
				target[p] = prop;
				if (typeof prop == 'function')
					prop._method = { clazz: clazz.class2, name: p };
			}
	},

	_redefineClass: function(name, superclass, constructor, methods, staticMethods) {
		var rn = name == null ? null : jasq._resolveName(name);
		constructor.class2 = new jasq._Class(name, constructor, superclass);

		if (typeof methods == 'object' && methods != null)
			jasq._inject(constructor, methods);
		if (typeof staticMethods == 'object' && staticMethods != null)
			jasq._injectStatic(constructor, staticMethods);

		if (rn != null)
			rn.base[rn.target] = constructor;
	},

	defineClass: function(name, superclass, constructor, methods, staticMethods) {
		if (typeof constructor != 'function')
			constructor = function() { };
		if (typeof superclass != 'function')
			superclass = Object;
		constructor.prototype = new superclass();
		constructor.prototype.constructor = constructor;

		jasq._redefineClass(name, superclass, constructor, methods, staticMethods);
		methods = staticMethods = null;	// Free up references
		return constructor;
	},

	defineThrowable: function(name) {
		java.lang.Throwable._lock = true;
		jasq.defineClass(name, java.lang.Throwable,
			function(message, cause) {
				java.lang.Throwable.call(this, message, cause);
			}
		);
		java.lang.Throwable._lock = false;
	},

	unsupportedOperation: function() {
		return function() {
			throw new java.lang.UnsupportedOperationException();
		};
	}
};

/* Package java.lang */

package2('java.lang');

jasq._redefineClass('java.lang.Boolean', /* extends */ Object,
	Boolean,
	{
		hashCode: function() {
			return this.valueOf() ? 1231 : 1237;
		},

		equals: function(obj) {
			obj = Object(obj);
			return obj instanceof Boolean && obj.valueOf() == this.valueOf();
		},

		compareTo: function(b) {
			if (this == b) return 0;
			else if (this) return 1;
			else return -1;
		}
	}
);

jasq._redefineClass('java.lang.Number', /* extends */ Object,
	Number,
	{
		hashCode: function() {
			var v = this.valueOf();
			return v ^ (v / 0x100000000);
		},

		equals: function(obj) {
			obj = Object(obj);
			return obj instanceof Number && obj.valueOf() == this.valueOf();
		},

		compareTo: function(a) {
			if (isNaN(this) || this > a) return 1;
			else if (this == a) return 0;
			else return -1;
		}
	}
);

jasq._redefineClass('java.lang.Object', /* extends */ null,
	Object,
	{
		hashCode: function() {
			return 0;
		},

		equals: function(obj) {
			return this === obj;
		},

		getClass: function() {
			return this.constructor.class2;
		},

		toString: function() {
			return this.getClass().getName();
		}
	}
);

jasq._redefineClass('java.lang.String', /* extends */ Object,
	String,
	{
		isEmpty: function() {
			return this.length == 0;
		},

		hashCode: function() {
			var h = 0;
			for (var i = 0; i < this.length; i++)
				h = (h * 31 + this.charCodeAt(i)) | 0;
			return h;
		},

		equals: function(obj) {
			obj = Object(obj);
			return obj instanceof String && obj.toString() == this.toString();
		},

		contentEquals: function(obj) {
			return obj.toString() == this.toString();
		},

		equalsIgnoreCase: function(str) {
			return this.toLowerCase() == str.toLowerCase();
		},

		codePointAt: String.prototype.charCodeAt,

		compareTo: function(str) {
			if (this > str)
				return 1;
			else if (this == str)
				return 0;
			else
				return -1;
		},

		compareToIgnoreCase: function(str) {
			return this.toLowerCase().compareTo(str.toLowerCase());
		},

		regionMatches: function(ignoreCase, toffset, other, ooffset, len) {
			if (typeof len == 'undefined') {
				len = ooffset;
				ooffset = other;
				other = toffset;
				toffset = ignoreCase;
				ignoreCase = false;
			}
			if (ignoreCase)
				return this.toLowerCase().substr(toffset, len) == other.toLowerCase().substr(ooffset, len);
			else
				return this.substr(toffset, len) == other.substr(ooffset, len);
		},

		startsWith: function(prefix, toffset) {
			if (typeof toffset != 'undefined')
				return this.substring(toffset).indexOf(prefix) == 0;
			else
				return this.indexOf(prefix) == 0;
		},

		endsWith: function(suffix) {
			return this.lastIndexOf(suffix) == this.length - suffix.length;
		},

		subSequence: function(begin, end) {
			return this.substring(begin, end);
		},

		contains: function(s) {
			return this.indexOf(s.toString()) > -1;
		},

		matches: function(regex) {
			return (new RegExp(regex)).test(this);
		},

		trim: function() {
			if (this.length == 0)
				return this;

			var start = 0;
			while (this.charCodeAt(start) <= 0x20)
				start++;

			var end = this.length - 1;
			while (this.charCodeAt(end) <= 0x20)
				end--;

			return this.substring(start, end + 1);
		}
	}
);

Array.class2 = new jasq._Class('[]', Array, Object);

jasq._redefineClass('java.lang.Math', /* extends */ Object,
	Math,
	null,
	{
		_prng:		null,
		_random:	Math.random,

		toRadians: function(angdeg) {
			return angdeg * Math.PI / 180;
		},

		toDegrees: function(angrad) {
			return angrad * 180 / Math.PI;
		},

		log10: function(a) {
			return Math.log(a) / Math.LN10;
		},

		cbrt: function(a) {
			return Math.pow(a, 1/3);
		},

		IEEEremainder: function(f1, f2) {
			return f1 - f2 * Math.rint(f1 / f2);
		},

		rint: function(a) {
			var f = Math.floor(a),
			    r = a - f;
			if (r == 0.5)
				return (f & 1) == 0 ? f : f + 1;
			else if (r > 0.5)
				return f + 1;
			else
				return f;
		},

		random: function() {
			if (Math._prng == null) {
				if (!(java.util instanceof Package))
					return Math._random();
				Math._prng = new java.util.Random();
			}
			return Math._prng.nextDouble();
		},

		signum: function(d) {
			if (S.isNaN(d))
				return d;
			else if (d == 0)
				return 0;
			else if (d > 0)
				return 1;
			else
				return -1;
		},

		sinh: function(x) {
			return (Math.exp(x) - Math.exp(-x)) / 2;
		},

		cosh: function(x) {
			return (Math.exp(x) + Math.exp(-x)) / 2;
		},

		tanh: function(x) {
			var e1 = Math.exp(x),
			    e2 = Math.exp(-x);
			return (e1 - e2) / (e1 + e2);
		},

		hypot: function(x, y) {
			return Math.sqrt(x*x + y*y);
		}
	}
);

jasq._redefineClass('java.lang.Class', /* extends */ Object,
	jasq._Class,
	{
		toString: function() {
			return this.clazz == null ? this.name : 'class ' + this.name;
		},

		newInstance: function() {
			try {
				return new this.clazz();
			} catch (ex) {
				throw new java.lang.InstantiationError(this.name);
			}
		},

		isInstance: function(obj) {
			return obj instanceof this.clazz;
		},

		isAssignableFrom: function(cls) {
			if (this.clazz == null)
				return this === cls;

			for (;;) {
				if (this === cls)
					return true;
				else if (cls.superclazz == null)
					return false;
				else
					cls = cls.superclazz.class2;
			}
		},

		isArray: function() {
			return this.clazz === Array;
		},

		isPrimitive: function() {
			return this.clazz == null;
		},

		getName: function() {
			return this.clazz === Array ? '[' : this.name;
		},

		getSuperclass: function() {
			return this.superclazz == null ? null : this.superclazz.class2;
		},

		getPackage: function() {
			try {
				return jasq._resolveName(this.name).base;
			} catch (ex) {
				throw new java.lang.InternalError();
			}
		}
	},
	{
		forName: function(className) {
			try {
				var rn = jasq._resolveName(className),
				    clazz = rn.base[rn.target];
				if (typeof clazz != 'function' || !(clazz.class2 instanceof java.lang.Class))
					throw null;
				return clazz.class2;
			} catch (ex) {
				throw new java.lang.ClassNotFoundException(className);
			}
		}
	}
);

jasq._redefineClass('java.lang.Package', /* extends */ Object,
	jasq._Package,
	{
		getName: function() {
			return this.name;
		},

		hashCode: function() {
			return this.name.hashCode();
		},

		toString: function() {
			return 'package ' + this.name;
		}
	},
	{
		getPackage: function(name) {
			try {
				var rn = jasq._resolveName(name + '.*');
				return rn.base;
			} catch (ex) {
				return null;
			}
		}
	}
);

jasq.defineClass('java.lang.StackTraceElement', /* extends */ null,
	/* constructor */ function(declaringClass, methodName, fileName, lineNumber) {
		this.clazz = declaringClass;
		this.method = methodName;
		this.file = fileName;
		this.line = lineNumber;
	},
	/* methods */ {
		getFileName: function() {
			return this.file;
		},

		getLineNumber: function() {
			return this.line;
		},

		getClassName: function() {
			return this.clazz;
		},

		getMethodName: function() {
			return this.method;
		},

		isNativeMethod: function() {
			return this.line == -2;
		},

		toString: function() {
			return this.clazz + '.' + this.method;
			/*
			var str = this.clazz + '.' + this.method + '(';
			if (this.file != null) {
				str += this.file;
				if (this.line >= 0)
					str += ':' + this.line;
			} else if (this.line == -2)
				str += 'Native Method';
			else
				str += 'Unknown Source';
			return str + ')';
			*/
		},

		equals: function(obj) {
			return obj instanceof java.lang.StackTraceElement && this.clazz == obj.clazz && this.method == obj.method && this.file == obj.file && this.line == obj.line;
		},

		hashCode: function() {
			return this.clazz.hashCode() + this.method.hashCode();
		}
	}
);

jasq.defineClass('java.lang.Throwable', /* extends */ null,
	/* constructor */ function(message, cause) {
		if (java.lang.Throwable._lock)
			return;

		if (message instanceof java.lang.Throwable) {
			cause = message;
			message = null;
		} else if (typeof message == 'undefined')
			message = null;

		this.message = message;
		this.cause = cause;
		this.fillInStackTrace();
	},
	/* methods */ {
		getMessage: function() {
			return this.message;
		},

		getLocalizedMessage: function() {
			return this.message;
		},

		getCause: function() {
			return typeof this.cause == 'undefined' ? null : this.cause;
		},

		initCause: function(cause) {
			if (cause == this)
				throw new java.lang.IllegalArgumentException();
			else if (typeof cause != 'undefined')
				throw new java.lang.IllegalStateException();
			this.cause = cause;
		},

		toString: function() {
			var name = this.getClass().getName(),
			    msg = this.getLocalizedMessage();
			return msg == null ? name : name + ': ' + msg;
		},

		fillInStackTrace: function() {
			if (typeof this.stackTrace != 'undefined')
				return this;

			var map = [],
			    start = -1,
			    caller = arguments.callee.caller,
			    thisClass = this.getClass();

			this.stackTrace = [];

			while (caller) {
				var e, str;

				if (caller.class2 instanceof java.lang.Class) {
					if (caller.class2 == thisClass)
						start = 0;
					e = new java.lang.StackTraceElement(caller.class2.getName(), '<init>', null, -1);
				} else if (typeof caller._method == 'object' && caller._method.clazz instanceof java.lang.Class)
					e = new java.lang.StackTraceElement(caller._method.clazz.getName(), caller._method.name, null, -1);
				else
					e = new java.lang.StackTraceElement('(Unknown Class)', '(Unknown Method)', null, -1);

				str = e.toString();
				if (map[str])
					break;
				else
					map[str] = true;

				if (start >= 0 && start++ > 0)
					this.stackTrace.push(e);

				caller = caller.caller;
			}

			return this;
		},

		printStackTrace: function() {
			var str = 'jasq exception:\n' + this.toString();
			if (typeof this.stackTrace != 'undefined' && this.stackTrace.length > 0)
				str += '\n    at ' + this.stackTrace.join('\n    at ');
			alert(str);
		},

		getStackTrace: function() {
			return this.stackTrace;
		},

		setStackTrace: function(stackTrace) {
			this.stackTrace = stackTrace;
		}
	}
);

jasq._redefineClass('Error', Object,
	Error,
	java.lang.Throwable.prototype
);
jasq._inject(
	Error,
	{
		getClass: function() {
			return Error.class2;
		}
	}
);

/* java.lang exception classes */
jasq.defineThrowable('java.lang.Exception');
jasq.defineThrowable('java.lang.ClassNotFoundException');
jasq.defineThrowable('java.lang.RuntimeException');
jasq.defineThrowable('java.lang.IllegalArgumentException');
jasq.defineThrowable('java.lang.IllegalStateException');
jasq.defineThrowable('java.lang.IndexOutOfBoundsException');
jasq.defineThrowable('java.lang.NegativeArraySizeException');
jasq.defineThrowable('java.lang.UnsupportedOperationException');

import2('java.lang.*');

/* java.lang error classes */
jasq.defineThrowable('java.lang.Error');
jasq.defineThrowable('java.lang.InstantiationError');
jasq.defineThrowable('java.lang.InternalError');

