/*
 * Jasq platform for JavaScript
 * http://lmadhavan.com/software/jasq/
 *
 * Copyright 2010 L. Madhavan
 * Licensed under the BSD license.
 * See the license.txt file for licensing terms.
 *
 * Last modified on 31 August 2010
 */

package2('java.util');

jasq.defineClass('java.util.AbstractCollection', /* extends */ null,
	/* constructor */ null,
	/* methods */ {
		iterator:	jasq.unsupportedOperation(),
		size:		jasq.unsupportedOperation(),
		add:		jasq.unsupportedOperation(),

		isEmpty: function() {
			return this.size() == 0;
		},

		contains: function(o) {
			var it = this.iterator();
			while (it.hasNext()) {
				var e = it.next();
				if (o == null ? e == null : o.equals(e))
					return true;
			}
			return false;
		},

		toArray: function(a) {
			if (typeof a == 'undefined')
				a = [];
			else
				a.length = 0;

			var it = this.iterator();
			while (it.hasNext())
				a.push(it.next());
			return a;
		},

		remove: function(o) {
			if (o instanceof Object) {
				var it = this.iterator();
				while (it.hasNext()) {
					var e = it.next();
					if (o == null ? e == null : o.equals(e)) {
						it.remove();
						return true;
					}
				}
				return false;
			} else
				Function.unsupported();
		},

		containsAll: function(c) {
			var it = c.iterator();
			while (it.hasNext())
				if (!this.contains(it.next()))
					return false;
			return true;
		},

		addAll: function(c) {
			var it = c.iterator(),
			    result = false;
			while (it.hasNext())
				result = result || this.add(it.next());
			return result;
		},

		removeAll: function(c) {
			var it = this.iterator(),
			    result = false;
			while (it.hasNext())
				if (c.contains(it.next())) {
					it.remove();
					result = true;
				}
			return result;
		},

		retainAll: function(c) {
			var it = this.iterator(),
			    result = false;
			while (it.hasNext())
				if (!c.contains(it.next())) {
					it.remove();
					result = true;
				}
			return result;
		},

		clear: function() {
			var it = this.iterator();
			while (it.hasNext())
				it.remove();
		},

		toString: function() {
			return '[' + this.toArray().join(', ') + ']';
		}
	}
);

jasq.defineClass('java.util.AbstractList', /* extends */ java.util.AbstractCollection,
	/* constructor */ null,
	/* methods */ {
		get:		jasq.unsupportedOperation(),
		set:		jasq.unsupportedOperation(),

		indexOf: function(o) {
			var it = this.listIterator();
			while (it.hasNext()) {
				var e = it.next();
				if (o == null ? e == null : o.equals(e))
					return it.nextIndex() - 1;
			}
			return -1;
		},

		lastIndexOf: function(o) {
			var it = this.listIterator(this.size());
			while (it.hasPrevious()) {
				var e = it.previous();
				if (o == null ? e == null : o.equals(e))
					return it.nextIndex() - 1;
			}
			return -1;
		},

		clear: function() {
			this._removeRange(0, this.size());
		},

		addAll: function(index, c) {
			if (typeof c == 'undefined') {
				c = index;
				index = this.size();
			}

			var it = c.iterator(),
			    result = false;
			while (it.hasNext())
				result = result || this.add(index++, it.next());
			return result;
		},

		iterator: function() {
			return this.listIterator();
		},

		listIterator: function(index) {
			if (typeof index == 'undefined')
				index = 0;

			if (index < 0 || index > this.size())
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.size());

			var self = this,
			    curIndex = -1;
			return {
				hasNext: function() {
					return index < self.size();
				},

				next: function() {
					if (index >= self.size())
						throw new java.util.NoSuchElementException();

					var e = self.get(index);
					curIndex = index++;
					return e;
				},

				hasPrevious: function() {
					return index > 0;
				},

				previous: function() {
					if (index <= 0)
						throw new java.util.NoSuchElementException();

					var e = self.get(--index);
					curIndex = index;
					return e;
				},

				nextIndex: function() {
					return index;
				},

				previousIndex: function() {
					return index - 1;
				},

				remove: function() {
					if (curIndex == -1)
						throw new java.lang.IllegalStateException();

					self.remove(curIndex);
					index = curIndex;
					curIndex = -1;
				},

				set: function(o) {
					if (curIndex == -1)
						throw new java.lang.IllegalStateException();

					self.set(curIndex, o);
				},

				add: function(o) {
					self.add(index, o);
					index++;
					curIndex = -1;
				}
			};
		},

		equals: function(o) {
			if (o == this)
				return true;

			if (!(o instanceof AbstractList))
				return false;

			if (this.size() != o.size())
				return false;

			var it1 = this.iterator(),
			    it2 = o.iterator();
			while (it1.hasNext()) {
				var e1 = it1.next(),
				    e2 = it2.next();
				if (!(e1 == null ? e2 == null : e1.equals(e2)))
					return false;
			}
			return true;
		},

		hashCode: function() {
			var h = 1,
			    it = this.iterator();
			while (it.hasNext()) {
				var e = it.next();
				h = (31 * h + (e == null ? 0 : e.hashCode())) | 0;
			}
			return h;
		},

		_removeRange: function(from, to) {
			var it = this.listIterator(from);
			while (to > from) {
				it.next();
				it.remove();
			}
		}
	}
);

jasq.defineClass('java.util.AbstractMap', /* extends */ null,
	/* constructor */ null,
	/* methods */ {
		put:		jasq.unsupportedOperation(),
		entrySet:	jasq.unsupportedOperation(),

		size: function() {
			return this.entrySet().size();
		},

		isEmpty: function() {
			return this.size() == 0;
		},

		containsValue: function(value) {
			var it = this.entrySet().iterator();
			while (it.hasNext()) {
				var v = it.next().getValue();
				if (value == null ? v == null : value.equals(v))
					return true;
			}
			return false;
		},

		containsKey: function(key) {
			var it = this.entrySet().iterator();
			while (it.hasNext()) {
				var k = it.next().getKey();
				if (key == null ? k == null : key.equals(k))
					return true;
			}
			return false;
		},

		get: function(key) {
			var it = this.entrySet().iterator();
			while (it.hasNext()) {
				var e = it.next(),
				    k = e.getKey();
				if (key == null ? k == null : key.equals(k))
					return e.getValue();
			}
			return null;
		},

		remove: function(key) {
			var it = this.entrySet().iterator();
			while (it.hasNext()) {
				var e = it.next(),
				    k = e.getKey();
				if (key == null ? k == null : key.equals(k)) {
					var v = e.getValue();
					it.remove();
					return v;
				}
			}
			return null;
		},

		putAll: function(t) {
			var it = t.entrySet().iterator();
			while (it.hasNext()) {
				var e = it.next();
				this.put(e.getKey(), e.getValue());
			}
		},

		clear: function() {
			this.entrySet().clear();
		},

		equals: function(o) {
			if (o == this)
				return true;

			if (!(o instanceof AbstractMap))
				return false;

			if (this.size() != o.size())
				return false;

			var it1 = this.entrySet().iterator(),
			    it2 = o.entrySet().iterator();
			while (it1.hasNext()) {
				var e1 = it1.next(),
				    e2 = it2.next(),
				    k1 = e1.getKey(),
				    k2 = e2.getKey(),
				    v1 = e1.getValue(),
				    v2 = e2.getValue();
				if (!(k1 == null ? k2 == null : k1.equals(k2)) ||
				    !(v1 == null ? v2 == null : v1.equals(v2)))
					return false;
			}
			return true;
		},

		hashCode: function() {
			var h = 0,
			    it = this.entrySet().iterator();
			while (it.hasNext())
				h = (h + it.next().hashCode()) | 0;
			return h;
		},

		toString: function() {
			var a = [],
			    it = this.entrySet().iterator();
			while (it.hasNext()) {
				var e = it.next();
				a.push(e.getKey() + '=' + e.getValue());
			}
			return '{' + a.join(', ') + '}';
		}
	}
);

jasq.defineClass('java.util.AbstractSequentialList', /* extends */ java.util.AbstractList,
	/* constructor */ null,
	/* methods */ {
		get: function(index) {
			var it = this.listIterator(index);
			return it.next();
		},

		set: function(index, element) {
			var it = this.listIterator(index);
			it.next();
			it.set(element);
		},

		add: function(index, element) {
			if (typeof element == 'undefined') {
				element = index;
				index = this.size();
			}

			var it = this.listIterator(index);
			it.add(element);
		},

		remove: function(index) {
			if (index instanceof Object)
				return java.util.AbstractCollection.prototype.remove.call(this, index);
			
			var it = this.listIterator(index),
			    old = it.next();
			it.remove();
			return old;
		},

		addAll: function(index, c) {
			if (typeof c == 'undefined') {
				c = index;
				index = this.size();
			}

			var it = this.listIterator(index),
			    cit = c.iterator();
			while (cit.hasNext()) {
				it.add(cit.next());
				it.next();
			}
		}
	}
);

jasq.defineClass('java.util.AbstractSet', /* extends */ java.util.AbstractCollection,
	/* constructor */ null,
	/* methods */ {
		equals: function(o) {
			return o instanceof java.util.AbstractSet && o.size() == this.size() && this.containsAll(o);
		},

		hashCode: function() {
			var h = 0,
			    it = this.iterator();
			while (it.hasNext())
				h = (h + it.next().hashCode()) | 0;
			return h;
		},

		removeAll: function(c) {
			if (this.size() < c.size()) {
				var it = this.iterator(),
				    result = false;
				while (it.hasNext())
					if (c.contains(it.next())) {
						it.remove();
						result = true;
					}
				return result;
			} else {
				var it = c.iterator(),
				    result = false;
				while (it.hasNext())
					result = result || this.remove(it.next());
				return result;
			}
		}
	}
);

jasq.defineClass('java.util.ArrayList', /* extends */ java.util.AbstractList,
	/* constructor */ function(c) {
		this.array = [];

		if (c instanceof AbstractCollection)
			this.addAll(c);
	},
	/* methods */ {
		trimToSize: function() {
			/* nothing to do; JS arrays are already dynamic */
		},

		ensureCapacity: function() {
			/* nothing to do; JS arrays are already dynamic */
		},

		size: function() {
			return this.array.length;
		},

		contains: function(o) {
			return this.indexOf(o) > -1;
		},

		indexOf: function(o) {
			for (var i = 0; i < this.array.length; i++) {
				var e = this.array[i];
				if (o == null ? e == null : o.equals(e))
					return i;
			}
			return -1;
		},

		lastIndexOf: function(o) {
			for (var i = this.array.length - 1; i >= 0; i--) {
				var e = this.array[i];
				if (o == null ? e == null : o.equals(e))
					return i;
			}
			return -1;
		},

		clone: function() {
			var a = new java.util.ArrayList();
			for (var i = 0; i < this.array.length; i++)
				a.array.push(this.array[i]);
			return a;
		},

		toArray: function(a) {
			if (typeof a == 'undefined')
				a = [];
			else
				a.length = 0;

			for (var i = 0; i < this.array.length; i++)
				a.push(this.array[i]);
			return a;
		},

		get: function(index) {
			if (index < 0 || index >= this.array.length)
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.array.length);

			return this.array[index];
		},

		set: function(index, element) {
			if (index < 0 || index >= this.array.length)
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.array.length);

			var old = this.array[index];
			this.array[index] = element;
			return old;
		},

		add: function(index, element) {
			if (typeof element == 'undefined') {
				element = index;
				this.array.push(element);
				return true;
			} else {
				if (index < 0 || index > this.array.length)
					throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.array.length);

				this.array.splice(index, 0, element);
			}
		},

		remove: function(index) {
			if (index instanceof Object)
				return java.util.AbstractCollection.prototype.remove.call(this, index);

			if (index < 0 || index >= this.array.length)
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.array.length);

			var old = this.array[index];
			this.array.splice(index, 1);
			return old;
		},

		clear: function() {
			this.array.length = 0;
		},

		addAll: function(index, c) {
			var size = this.array.length;

			if (typeof c == 'undefined') {
				c = index;
				index = size;
			} else if (index < 0 || index > size)
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.array.length);

			var count = c.size();
			if (count == 0)
				return false;

			for (var i = size - 1; i >= index; i--)
				this.array[i + count] = this.array[i];

			var it = c.iterator(),
			    i = index;
			while (it.hasNext())
				this.array[index++] = it.next();
			return true;
		},

		_removeRange: function(from, to) {
			this.array.splice(from, to - from);
		}
	}
);

jasq.defineClass('java.util.Arrays', /* extends */ null,
	/* constructor */ null,
	/* methods */ null,
	/* static methods */ {
		sort: function(a, from, to, c) {
			if (from instanceof Object) {
				c = from;
				from = void 0;
			}

			if (typeof c == 'undefined')
				c = function(e1, e2) {
					return e1.compareTo(e2);
				};
			else
				c = c.compare;

			if (typeof from == 'undefined' || (from == 0 && to == a.length)) {
				a.sort(c);
				return;
			}

			if (from == to)
				return;
			if (from > to)
				throw new java.lang.IllegalArgumentException();
			if (from < 0 || to > a.length)
				throw new java.lang.IndexOutOfBoundsException();

			var b = [];
			for (var i = from; i < to; i++)
				b.push(a[i]);
			b.sort(c);
			for (var i = from; i < to; i++)
				a[i] = b[i - from];
		},

		binarySearch: function(a, from, to, key, c) {
			switch (arguments.length) {
				case 2: key = from, from = void 0; break;
				case 3: key = from, c = to, from = void 0; break;
			}

			if (typeof c == 'undefined')
				c = function(e1, e2) {
					return e1.compareTo(e2);
				};
			else
				c = c.compare;

			if (typeof from == 'undefined')
				from = 0, to = a.length;
			if (from == to)
				return -from - 1;
			if (from > to)
				throw new java.lang.IllegalArgumentException();
			if (from < 0 || to > a.length)
				throw new java.lang.IndexOutOfBoundsException();

			var begin = from,
			    end = to - 1;
			while (begin <= end) {
				var mid = (begin + end) >> 1,
				    e = a[mid],
				    d = c(key, e);

				if (d == 0)
					return mid;
				else if (d < 0)
					end = mid - 1;
				else
					begin = mid + 1;
			}
			return -begin - 1;
		},

		equals: function(a1, a2) {
			if (a1 == null || a2 == null)
				return a1 == a2;

			if (a1.length != a2.length)
				return false;

			for (var i = 0; i < a1.length; i++)
				if (!a1[i].equals(a2[i]))
					return false;
			return true;
		},

		fill: function(a, from, to, val) {
			if (typeof val == 'undefined') {
				val = from;
				from = 0, to = a.length;
			}

			if (from == to)
				return;
			if (from > to)
				throw new java.lang.IllegalArgumentException();
			if (from < 0 || to > a.length)
				throw new java.lang.IndexOutOfBoundsException();

			for (var i = from; i < to; i++)
				a[i] = val;
		},

		copyOf: function(original, newLength) {
			if (newLength < 0)
				throw new java.lang.NegativeArraySizeException();

			var a = new Array(newLength),
			    n = Math.min(original.length, a.length);
			for (var i = 0; i < n; i++)
				a[i] = original[i];
			for (; i < a.length; i++)
				a[i] = 0;

			return a;
		},

		copyOfRange: function(original, from, to) {
			if (from > to)
				throw new java.lang.IllegalArgumentException();
			if (from < 0 || to > original.length)
				throw new java.lang.IndexOutOfBoundsException();

			var a = new Array(to - from),
			    n = Math.min(original.length, a.length);
			for (var i = 0; i < n; i++)
				a[i] = original[from + i];
			for (; i < a.length; i++)
				a[i] = 0;

			return a;
		},

		hashCode: function(a) {
			if (a == null)
				return 0;

			var h = 0;
			for (var i = 0; i < a.length; i++) {
				var e = a[i];
				h = (31 * h + (e == null ? 0 : e.hashCode())) | 0;
			}
			return h;
		},

		deepHashCode: function deepHashCode(a) {
			if (a == null)
				return 0;

			var h = 0;
			for (var i = 0; i < a.length; i++) {
				var e = a[i];
				h = (31 * h + (e == null ? 0 : (e instanceof Array ? deepHashCode(a[i]) : a[i].hashCode()))) | 0;
			}
			return h;
		},

		deepEquals: function deepEquals(a1, a2) {
			if (a1 == null || a2 == null)
				return a1 == a2;

			if (a1.length != a2.length)
				return false;

			for (var i = 0; i < a1.length; i++) {
				var e1 = a1[i],
				    e2 = a2[i];
				if (e1 instanceof Array) {
					if (!(e2 instanceof Array) || !deepEquals(e1, e2))
						return false;
				} else if (!a1[i].equals(a2[i]))
					return false;
			}

			return true;
		},

		toString: function(a) {
			return a == null ? 'null' : '[' + a.join(', ') + ']';
		},

		deepToString: function deepToString(a) {
			if (a == null)
				return 'null';

			var str = '[';
			for (var i = 0; i < a.length - 1; i++) {
				var e = a[i];
				str += (e == null ? 'null' : (e instanceof Array ? deepToString(e) : e.toString())) + ', ';
			}
			var e = a[a.length - 1];
			str += (e == null ? 'null' : (e instanceof Array ? deepToString(e) : e.toString()));
			return str + ']';
		}
	}
);

jasq.defineClass('java.util.Collections', /* extends */ null,
	/* constructor */ null,
	/* methods */ null,
	/* static methods */ {
		sort: function(list, c) {
			if (typeof c == 'undefined' || c == null)
				c = function(e1, e2) {
					return e1.compareTo(e2);
				}
			else
				c = c.compare;

			var a = list.toArray();
			a.sort(c);

			var i = 0,
			    it = list.listIterator();
			while (it.hasNext()) {
				it.next();
				it.set(a[i++]);
			}
		},

		binarySearch: function(list, key, c) {
			if (typeof c == 'undefined' || c == null)
				c = function(e1, e2) {
					return e1.compareTo(e2);
				}
			else
				c = c.compare;

			var begin = 0,
			    end = list.size() - 1;
			while (begin <= end) {
				var mid = (begin + end) >> 1,
				    e = list.get(mid),
				    d = c(key, e);

				if (d == 0)
					return mid;
				else if (d < 0)
					end = mid - 1;
				else
					begin = mid + 1;
			}
			return -begin - 1;
		},

		reverse: function(list) {
			var n = list.size();
			
			var fit = list.listIterator(0),
			    rit = list.listIterator(n);

			n /= 2;
			for (var i = 0; i < n; i++) {
				var l = fit.next(),
				    r = rit.previous();
				fit.set(r);
				rit.set(l);
			}
		},

		shuffle: function(list, rnd) {
			if (typeof rnd == 'undefined')
				rnd = new java.util.Random();

			if (list instanceof AbstractSequentialList) {
				var a = list.toArray();
				for (var i = a.length - 1; i > 0; i--) {
					var j = rnd.nextInt(i + 1),
					    ej = a[j];
					a[j] = a[i];
					a[i] = ej;
				}

				var it = list.listIterator();
				i = 0;
				while (it.hasNext()) {
					it.next();
					it.set(a[i++]);
				}
			} else {
				for (var i = list.size() - 1; i > 0; i--) {
					var j = rnd.nextInt(i + 1),
					    ei = list.get(i),
					    ej = list.get(j);
					list.set(i, ej);
					list.set(j, ei);
				}
			}
		},

		swap: function(list, i, j) {
			var iit = list.listIterator(i),
			    jit = list.listIterator(j);

			var ei = iit.next(),
			    ej = jit.next();
			iit.set(ej);
			jit.set(ei);
		},

		fill: function(list, obj) {
			var it = list.listIterator();
			while (it.hasNext()) {
				it.next();
				it.set(obj);
			}
		},

		copy: function(dest, src) {
			var dit = dest.listIterator(),
			    sit = src.listIterator();
			while (sit.hasNext()) {
				dit.next();
				dit.set(sit.next());
			}
		},

		min: function(coll, comp) {
			if (coll.size() == 0)
				throw new java.lang.IndexOutOfBoundsException('Collection is empty.');

			if (typeof comp == 'undefined' || comp == null)
				comp = function(e1, e2) {
					return e1.compareTo(e2);
				}
			else
				comp = comp.compare;

			var it = coll.iterator(),
			    m = it.next();
			while (it.hasNext()) {
				var e = it.next();
				if (comp(e, m) < 0)
					m = e;
			}
			return m;
		},

		max: function(coll, comp) {
			if (coll.size() == 0)
				throw new java.lang.IndexOutOfBoundsException('Collection is empty.');

			if (typeof comp == 'undefined' || comp == null)
				comp = function(e1, e2) {
					return e1.compareTo(e2);
				}
			else
				comp = comp.compare;

			var it = coll.iterator(),
			    m = it.next();
			while (it.hasNext()) {
				var e = it.next();
				if (comp(e, m) > 0)
					m = e;
			}
			return m;
		},

		replaceAll: function(list, oldVal, newVal) {
			var it = list.listIterator(),
			    result = false;
			while (it.hasNext()) {
				var e = it.next();
				if (oldVal == null ? e == null : oldVal.equals(e)) {
					e.set(newVal);
					result = true;
				}
			}
			return result;
		},

		reverseOrder: function(cmp) {
			if (typeof cmp == 'undefined' || cmp == null)
				return {
					compare: function(e1, e2) {
						return e2.compareTo(e1);
					}
				};
			else
				return {
					compare: function(e1, e2) {
						return cmp.compare(e2, e1);
					}
				};
		},

		frequency: function(c, o) {
			var it = c.iterator(),
			    freq = 0;
			while (it.hasNext()) {
				var e = it.next();
				if (o == null ? e == null : o.equals(e))
					freq++;
			}
			return freq;
		},

		disjoint: function(c1, c2) {
			if (c1.size() > c2.size()) {
				var c = c1;
				c1 = c2;
				c2 = c;
			}

			var it = c1.iterator();
			while (it.hasNext())
				if (c2.contains(it.next()))
					return false;
			return true;
		},

		addAll: function(c) {
			var result = false;
			for (var i = 1; i < arguments.length; i++)
				result = result || c.add(arguments[i]);
			return result;
		},

		_SetFromMap: jasq.defineClass('java.util.Collections$SetFromMap', java.util.AbstractSet,
			/* constructor */ function(m) {
				this.map = m;
			},
			/* methods */ {
				iterator: function() {
					return this.map.keySet().iterator();
				},

				size: function() {
					return this.map.size();
				},

				contains: function(o) {
					return this.map.containsKey(o);
				},

				add: function(o) {
					return this.map.put(o, true) != true;
				},

				remove: function(o) {
					return this.map.remove(o) == true;
				},

				clear: function() {
					this.map.clear();
				}
			}
		),

		newSetFromMap: function(map) {
			return new this._SetFromMap(map);
		}
	}
);

jasq.defineClass('java.util.HashMap', /* extends */ java.util.AbstractMap,
	/* constructor */ function(initialCapacity, loadFactor) {
		var m = null;
		if (initialCapacity instanceof java.util.AbstractMap) {
			m = initialCapacity;
			loadFactor = 0.75;
			initialCapacity = Math.ceil(m.size() / loadFactor);
		} else {
			if (typeof initialCapacity == 'undefined')
				initialCapacity = 16;
			else if (initialCapacity < 0)
				throw new java.lang.IllegalArgumentException('initial capacity: ' + initialCapacity);
			if (typeof loadFactor == 'undefined')
				loadFactor = 0.75;
			else if (loadFactor <= 0)
				throw new java.lang.IllegalArgumentException('load factor: ' + loadFactor);
		}

		this.buckets = [];
		this.loadFactor = loadFactor;
		this.length = 0;

		for (var i = 0; i < initialCapacity; i++)
			this.buckets[i] = [];

		if (m != null)
			this.putAll(m);
	},
	/* methods */ {
		size: function() {
			return this.length;
		},

		get: function(key) {
			var h = key.hashCode() % this.buckets.length,
			    bucket = this.buckets[h];

			for (var i = 0; i < bucket.length; i++) {
				var e = bucket[i];
				if (key == null ? e.key == null : key.equals(e.key))
					return e.value;
			}
			return null;
		},

		containsKey: function(key) {
			var h = key.hashCode() % this.buckets.length,
			    bucket = this.buckets[h];

			for (var i = 0; i < bucket.length; i++) {
				var e = bucket[i];
				if (key == null ? e.key == null : key.equals(e.key))
					return true;
			}
			return false;
		},

		containsValue: function(value) {
			for (var i = 0; i < this.buckets.length; i++) {
				var bucket = this.buckets[i];
				for (var i = 0; i < bucket.length; i++) {
					var v = bucket[i].value;
					if (value == null ? v == null : value.equals(v))
						return true;
				}
			}
			return false;
		},

		put: function(key, value) {
			if (this.buckets.length == 0)
				this.buckets = [[]];

			var h = key.hashCode() % this.buckets.length,
			    bucket = this.buckets[h];

			for (var i = 0; i < bucket.length; i++) {
				var e = bucket[i];
				if (key == null ? e.key == null : key.equals(e.key)) {
					var old = e.value;
					e.value = value;
					return old;
				}
			}

			bucket.push({ key: key, value: value });
			this.length++;

			if (this.length >= this.buckets.length * this.loadFactor)
				this._rehash();

			return null;
		},

		remove: function(key) {
			var h = key.hashCode() % this.buckets.length,
			    bucket = this.buckets[h];

			for (var i = 0; i < bucket.length; i++) {
				var e = bucket[i];
				if (key == null ? e.key == null : key.equals(e.key)) {
					bucket.splice(i, 1);
					return e.value;
				}
			}
			return null;
		},

		clear: function() {
			for (var i = 0; i < this.buckets.length; i++)
				this.buckets[i] = [];
			this.length = 0;
		},

		clone: function() {
			return new java.util.HashMap(this);
		},

		_KeySet: jasq.defineClass('java.util.HashMap$KeySet', /* extends */ java.util.AbstractSet,
			/* constructor */ function(map) {
				this.map = map;
			},
			/* methods */ {
				iterator: function() {
					var map = this.map,
					    buckets = map.buckets,
					    a = 0, b = -1, curA = -1, curB = -1;

					function findNext() {
						while (a < buckets.length) {
							if (++b < buckets[a].length)
								return;
							a++, b = -1;
						}
					}

					findNext();
					return {
						hasNext: function() {
							return a < buckets.length;
						},

						next: function() {
							if (a >= buckets.length)
								throw new java.lang.NoSuchElementException();

							var k = buckets[a][b].key;
							curA = a, curB = b;
							findNext();
							return k;
						},

						remove: function() {
							if (curA == -1)
								throw new java.lang.IllegalStateException();

							buckets[curA].splice(curB, 1);
							map.length--;
							a = curA, b = curB - 1;
							findNext();
							curA = -1;
						}
					};
				},

				remove: function(o) {
					return this.map.remove(o);
				},

				clear: function() {
					this.map.clear();
				}
			}
		),

		keySet: function() {
			return new this._KeySet(this);
		},

		_Values: jasq.defineClass('java.util.HashMap$Values', /* extends */ java.util.AbstractCollection,
			/* constructor */function(map) {
				this.map = map;
			},
			/* methods */ {
				iterator: function() {
					var map = this.map,
					    buckets = map.buckets,
					    a = 0, b = -1, curA = -1, curB = -1;

					function findNext() {
						while (a < buckets.length) {
							if (++b < buckets[a].length)
								return;
							a++, b = -1;
						}
					}

					findNext();
					return {
						hasNext: function() {
							return a < buckets.length;
						},

						next: function() {
							if (a >= buckets.length)
								throw new java.util.NoSuchElementException();

							var v = buckets[a][b].value;
							curA = a, curB = b;
							findNext();
							return v;
						},

						remove: function() {
							if (curA == -1)
								throw new java.lang.IllegalStateException();

							buckets[curA].splice(curB, 1);
							map.length--;
							a = curA, b = curB - 1;
							findNext();
							curA = -1;
						}
					};
				},

				clear: function() {
					this.map.clear();
				}
			}
		),

		values: function() {
			return new this._Values(this);
		},

		_MapEntry: jasq.defineClass('java.util.HashMap$MapEntry', /* extends */ null,
			/* constructor */ function(entry) {
				this.entry = entry;
			},
			/* methods */ {
				getKey: function() {
					return this.entry.key;
				},

				getValue: function() {
					return this.entry.value;
				},

				setValue: function(value) {
					this.entry.value = value;
				},

				hashCode: function() {
					return (this.entry.key == null ? 0 : this.entry.key.hashCode()) ^ (this.entry.value == null ? 0 : this.entry.value.hashCode());
				},

				equals: function(o) {
					return o instanceof MapEntry &&
					       (this.entry.key == null ? o.entry.key == null : this.entry.key.equals(o.entry.key)) &&
					       (this.entry.value == null ? o.entry.value == null : this.entry.value.equals(o.entry.value));
				}
			}
		),

		_EntrySet: jasq.defineClass('java.util.HashMap$EntrySet', /* extends */ java.util.AbstractSet,
			/* constructor */ function(map) {
				this.map = map;
			},
			/* methods */ {
				iterator: function() {
					var map = this.map,
					    buckets = map.buckets,
					    a = 0, b = -1, curA = -1, curB = -1;

					function findNext() {
						while (a < buckets.length) {
							if (++b < buckets[a].length)
								return;
							a++, b = -1;
						}
					}

					findNext();
					return {
						hasNext: function() {
							return a < buckets.length;
						},

						next: function() {
							if (a >= buckets.length)
								throw new java.util.NoSuchElementException();

							var e = buckets[a][b];
							curA = a, curB = b;
							findNext();
							return new map._MapEntry(e);
						},

						remove: function() {
							if (curA == -1)
								throw new java.lang.IllegalStateException();

							buckets[curA].splice(curB, 1);
							map.length--;
							a = curA, b = curB - 1;
							findNext();
							curA = -1;
						}
					};
				},

				clear: function() {
					this.map.clear();
				}
			}
		),

		entrySet: function() {
			return new this._EntrySet(this);
		},

		_rehash: function() {
			var oldBuckets = this.buckets;

			this.buckets = [];
			for (var i = 0; i < oldBuckets.length * 2; i++)
				this.buckets[i] = [];

			for (var i = 0; i < oldBuckets.length; i++) {
				var bucket = oldBuckets[i];
				for (var j = 0; j < bucket.length; j++) {
					var e = bucket[j],
					    h = e.key.hashCode() % this.buckets.length;
					this.buckets[h].push(e);
				}
			}
		}
	}
);

jasq.defineClass('java.util.HashSet', /* extends */ java.util.AbstractSet,
	/* constructor */ function(initialCapacity, loadFactor) {
		if (initialCapacity instanceof java.util.AbstractCollection) {
			var c = initialCapacity;
			this.map = new java.util.HashMap(Math.ceil(c.size() / 0.75), 0.75);
			this.addAll(c);
		} else
			this.map = new java.util.HashMap(initialCapacity, loadFactor);
	},
	/* methods */ {
		iterator: function() {
			return this.map.keySet().iterator();
		},

		size: function() {
			return this.map.size();
		},

		contains: function(o) {
			return this.map.containsKey(o);
		},

		add: function(o) {
			return this.map.put(o, true) != true;
		},

		remove: function(o) {
			return this.map.remove(o) == true;
		},

		clear: function() {
			this.map.clear();
		},

		clone: function() {
			return new java.util.HashSet(this);
		}
	}
);

jasq.defineClass('java.util.LinkedList', /* extends */ java.util.AbstractSequentialList,
	/* constructor */ function(c) {
		this.head = { prev: null, next: null };
		this.tail = { prev: null, next: null };
		this.clear();

		if (c instanceof AbstractCollection)
			this.addAll(c);
	},
	/* methods */ {
		getFirst: function() {
			if (this.length == 0)
				throw new java.lang.IndexOutOfBoundsException('List is empty.');
			return this.head.next.value;
		},

		getLast: function() {
			if (this.length == 0)
				throw new java.lang.IndexOutOfBoundsException('List is empty.');
			return this.tail.prev.value;
		},

		removeFirst: function() {
			if (this.length == 0)
				throw new java.lang.IndexOutOfBoundsException('List is empty.');

			var old = this.head.next;
			this.head.next = old.next;
			old.next.prev = this.head;
			this.length--;
			return old.value;
		},

		removeLast: function() {
			if (this.length == 0)
				throw new java.lang.IndexOutOfBoundsException('List is empty.');

			var old = this.tail.next;
			this.tail.prev = old.prev;
			old.prev.next = this.tail;
			this.length--;
			return old.value;
		},

		addFirst: function(o) {
			var node = { prev: this.head, next: this.head.next, value: o };
			node.prev.next = node;
			node.next.prev = node;
			this.length++;
		},

		addLast: function(o) {
			var node = { prev: this.tail.prev, next: this.tail, value: o };
			node.prev.next = node;
			node.next.prev = node;
			this.length++;
		},

		contains: function(o) {
			return this.indexOf(o) > -1;
		},

		indexOf: function(o) {
			if (this.length == 0)
				return false;

			var node, i;
			for (node = this.head.next, i = 0; node != this.tail; node = node.next, i++)
				if (o == null ? node.value == null : o.equals(node.value))
					return i;
			return -1;
		},

		lastIndexOf: function(o) {
			if (this.length == 0)
				return false;

			var node, i;
			for (node = this.tail.prev, i = this.length - 1; node != this.head; node = node.prev, i--)
				if (o == null ? node.value == null : o.equals(node.value))
					return i;
			return -1;
		},

		size: function() {
			return this.length;
		},

		clear: function() {
			this.head.next = this.tail;
			this.tail.prev = this.head;
			this.length = 0;
		},

		listIterator: function(index) {
			if (typeof index == 'undefined')
				index = 0;

			if (index < 0 || index > this.length)
				throw new java.lang.IndexOutOfBoundsException('index ' + index + ', size ' + this.size());

			var node;
			if (index < this.length / 2) {
				node = this.head.next;
				for (var i = 0; i < index; i++)
					node = node.next;
			} else {
				node = this.tail;
				for (var i = this.length - index; i > 0; i--)
					node = node.prev;
			}

			var self = this,
			    curNode = null;
			return {
				hasNext: function() {
					return node != self.tail;
				},

				next: function() {
					if (node == self.tail)
						throw new java.util.NoSuchElementException();
					
					curNode = node;
					node = node.next;
					index++;
					return curNode.value;
				},

				hasPrevious: function() {
					return node.prev != this.head;
				},

				previous: function() {
					if (node.prev == self.head)
						throw new java.util.NoSuchElementException();
					
					node = node.prev;
					curNode = node;
					index--;
					return node.value;
				},

				nextIndex: function() {
					return index;
				},

				previousIndex: function() {
					return index - 1;
				},

				remove: function() {
					if (curNode == null)
						throw new java.lang.IllegalStateException();

					curNode.prev.next = curNode.next;
					curNode.next.prev = curNode.prev;
					self.length--;
					node = curNode.next;
					curNode = null;
				},

				set: function(o) {
					if (curNode == null)
						throw new java.lang.IllegalStateException();

					curNode.value = o;
				},

				add: function(o) {
					var newNode = { prev: node.prev, next: node, value: o };
					node.prev.next = newNode;
					node.prev = newNode;
					self.length++;
					index++;
					curNode = null;
				}
			};	
		},

		clone: function() {
			return new java.util.LinkedList(this);
		},

		toArray: function(a) {
			if (typeof a == 'undefined')
				a = [];
			else
				a.length = 0;

			for (var node = this.head.next; node != this.tail; node = node.next)
				a.push(node.value);
			return a;
		}
	}
);

jasq.defineClass('java.util.Random', /* extends */ null,
	/* constructor */ function(seed) {
		if (typeof seed == 'undefined')
			seed = (new Date()).getTime();
		
		this.setSeed(seed);
	},
	/* methods */ {
		setSeed: function(seed) {
			this.seed = seed ^ 1664525;
			this.haveNextNextGaussian = false;
		},

		_next: function(bits) {
			this.seed = (this.seed * 1664525 + 1013904223) | 0;
			return this.seed >>> (32 - bits);
		},

		nextInt: function(n) {
			if (typeof n == 'undefined')
				return this._next(32);

			n |= 0;
			if (n <= 0)
				throw new java.lang.IllegalArgumentException('n: ' + n);

			var bits, val;
			do {
				bits = this._next(31);
				val = bits % n;
			} while (bits - val + (n-1) < 0);
			return val;
		},

		nextBoolean: function() {
			return this._next(1) != 0;
		},

		nextDouble: function() {
			return (this._next(26) * Math.pow(2, 27) + this._next(27)) / Math.pow(2, 53);
		},

		nextGaussian: function() {
			if (this.haveNextNextGaussian) {
				this.haveNextNextGaussian = false;
				return this.nextNextGaussian;
			} else {
				var v1, v2, s;
				do {
					v1 = 2 * this.nextDouble() - 1;
					v2 = 2 * this.nextDouble() - 1;
					s = v1 * v1 + v2 * v2;
				} while (s >= 1 || s == 0);

				var multiplier = Math.sqrt(-2 * Math.log(s) / s);
				this.nextNextGaussian = v2 * multiplier;
				this.haveNextNextGaussian = true;
				return v1 * multiplier;
			}
		}
	}
);

jasq.defineThrowable('java.util.NoSuchElementException');

